#!/usr/bin/python3

from typing import Dict, Iterable, Any, Optional
from configparser import ConfigParser
from datetime import datetime
import os, json, logging, logging.config

from bottle import route, run, response, request


servicePath = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


class DocumentsStore(object):
    """
    Хранилище чеков и информации по чекам
    """
    docsDir = os.path.join(servicePath, 'documents')
    allDocs = os.path.join(docsDir, 'alldocs.json')
    _instance = None

    def __new__(self, *args, **kwargs):
        if not self._instance:
            self._instance = super(DocumentsStore, self).__new__(self, *args, **kwargs)
            self.documents = {}
            if not os.path.isdir(self.docsDir):
                os.mkdir(self.docsDir)
            if os.path.isfile(self.allDocs):
                with open(self.allDocs) as f:
                    self.documents = json.load(f)
        return self._instance

    def docFileName(self, docId: str) -> str:
        """
        Возвращает полное имя файла чека на ФС по его ID
        """
        return os.path.join(self.docsDir, 'doc_%s.img' % docId)

    def saveDocMeta(self, data: Dict[str, Any]) -> None:
        """
        Сохранение метаинформации о документе
        """
        docId = data['documentid']
        docMeta = data.copy()
        if 'document' in docMeta:
            del docMeta['document']
        docMeta["updateTime"] = datetime.now().isoformat()
        self.documents[docId] = docMeta
        with open(self.allDocs, 'w') as f:
            f.write(json.dumps(self.documents, sort_keys=True, ensure_ascii=True, indent=2))

    def saveDocImg(self, docId: str, document: str) -> None:
        """
        Сохранение образа чека
        """
        with open(self.docFileName(docId), 'w') as f:
            f.write(document)

    def docMeta(self, docId: str) -> Dict[str, Any]:
        """
        Возвращает метаданные о документе
        """
        if not docId in self.documents or not os.path.isfile(self.docFileName(docId)):
            raise Exception(u"Документ с id='%s' не найден" % docId)
        return self.documents[docId].copy()

    def loadDoc(self, docId: str) -> Dict[str, Any]:
        """
        Восстанавливает документ и метаданные по ID
        """
        data = self.docMeta(docId)
        with open(self.docFileName(docId)) as f:
            data['document'] = f.read()
        return data

    def getDocIds(self) -> Iterable[str]:
        """
        Возвращает список id документов
        """
        return self.documents.keys()

def getResult(status=0, comment=u'OK') -> dict:
    return {u'errorCode': status, u'message': comment}

@route('/pendingchecks', method='POST')
def postDocument() -> Optional[str]:
    try:
        logging.info(u'Получили запрос на сохранение документа')
        logging.info(u'URL: %s' % request.url)
        body = json.loads(request.body.read())

        requiredFields = ['documentid', 'document', 'state']
        for i in requiredFields:
            if not i in body:
                raise Exception(u'Отсутствует обязательное поле %s' % i)

        store = DocumentsStore()
        store.saveDocImg(body['documentid'], body['document'])
        store.saveDocMeta(body)

    except Exception as e:
        logging.error(e)
        response.status = 400
        return json.dumps(getResult(400, str(e)), ensure_ascii=False)

@route('/pendingchecks/bypage', method='GET')
def getDocuments() -> str:
    try:
        logging.info(u'Запрос списка документов')
        logging.info(u'URL: %s' % request.url)

        documents = []
        store = DocumentsStore()
        states = request.query.getall('states')
        startDate = datetime.today()
        if (request.query.startdate):
            startDate = datetime.strptime(request.query.startdate, "%Y-%m-%d").date()
        for docId in store.getDocIds():
            if store.docMeta(docId)['state'] not in states:
                continue
            document = store.loadDoc(docId)
            if request.query.startdate:
                documentDate = datetime.strptime(store.docMeta(docId)['updateTime'], "%Y-%m-%dT%H:%M:%S.%f").date()
                if (documentDate < startDate):
                    continue
            documents.append(document)
        result = {}
        result["content"] = documents
        result["number"] = 0
        result["size"] = 100
        result["lastPage"] = True
        result["totalElements"] = len(documents)
        result["numberOfElements"] = len(documents)
        result["totalPages"] = 1
        result["firstPage"] = True
        return json.dumps(result, ensure_ascii=False)

    except Exception as e:
        logging.error(e)
        response.status = 400
        return json.dumps(getResult(400, str(e)), ensure_ascii=False)

@route('/pendingchecks/<docId>', method='GET')
def getDocument(docId: str) -> str:
    try:
        logging.info(u'Запрос документа %s' % docId)
        logging.info(u'URL: %s' % request.url)

        store = DocumentsStore()
        return json.dumps(store.loadDoc(docId), ensure_ascii=False)
    except Exception as e:
        logging.error(e)
        response.status = 400
        return json.dumps(getResult(400, str(e)), ensure_ascii=False)

@route('/pendingchecks/<docId>', method='PUT')
def updateDocument(docId: str) -> Optional[str]:
    try:
        logging.info(u'Изменение документа %s' % docId)
        logging.info(u'URL: %s' % request.url)
        body = json.loads(request.body.read())

        if not 'state' in body and not 'document' in body:
            raise Exception(u"В запросе должно присутствовать как минимум одно из полей - 'state' или 'document'")

        store = DocumentsStore()
        docMeta = store.docMeta(docId)

        if docMeta['state'] == body['state']:
            raise Exception(u'Документ уже находится в этом статусе, статус не может быть изменен.')

        if 'document' in body:
            store.saveDocImg(docId, body['document'])

        if 'state' in body:
            docMeta['state'] = body['state']
            store.saveDocMeta(docMeta)

    except Exception as e:
        logging.error(e)
        response.status = 400
        return json.dumps(getResult(400, str(e)), ensure_ascii=False)

def main():
    logging.basicConfig(format = u'[%(asctime)s] %(levelname)s - %(message)s', level = logging.INFO,
            filename = u'/linuxcash/logs/current/deferredchecktest.log')
    logging.info(u'Старт сервиса')
    try:
        conf = ConfigParser()
        conf.read(os.path.join(servicePath, 'conf/config.ini'))
        try:
            host = conf.get('main', 'host')
        except:
            host = "0.0.0.0"

        try:
            port = conf.getint('main', 'port')
        except:
            port = 18087

        logging.info(u'Стартуем на %s:%s' % (host, port))
        run(host=host, port=port)
    except Exception as e:
        logging.error(e)

if __name__ == '__main__':
    main()
